@@ -51,7 +51,6 @@ void net_login(const char *username, const char *password)
 	if(!port) port = 5688;
 	ggzcore_server_set_hostinfo(server, host, port, 0);
 	printf("Logging in as %s...\n", m_username);
-	fflush(NULL);
 	ggzcore_server_connect(server);
 }
 
@@ -74,7 +73,6 @@ void net_allow(int allow)
 	if(!allow)
 	{
 		printf("* Buffered input\n");
-		fflush(NULL);
 	}
 	m_allow = allow;
 }
@@ -87,12 +85,12 @@ void net_host(const char *hostname, int portnumber)
 	host = strdup(hostname);
 	port = portnumber;
 	printf("TelGGZ: Host is now %s:%i.\n", hostname, port);
-	fflush(NULL);
 }
 
 void net_join(int roomnum)
 {
-	if(room) ggzcore_server_join_room(server, roomnum);
+	GGZRoom *ggzroom = ggzcore_server_get_nth_room(server, roomnum);
+	if(room) ggzcore_server_join_room(server, ggzroom);
 }
 
 void net_list(void)
@@ -105,7 +103,6 @@ void net_list(void)
 		room = ggzcore_server_get_nth_room(server, i);
 		printf("%3i: %s\n", i, ggzcore_room_get_name(room));
 	}
-	fflush(NULL);
 }
 
 void net_who(void)
@@ -113,7 +110,6 @@ void net_who(void)
 	if(!room)
 	{
 		printf("Not in a room yet.\n");
-		fflush(NULL);
 		return;
 	}
 
@@ -132,7 +128,6 @@ static GGZHookReturn net_hook_players(unsigned int id,
 		player = ggzcore_room_get_nth_player(room, i);
 		printf(" * %s\n", ggzcore_player_get_name(player));
 	}
-	fflush(NULL);
 
 	return GGZ_HOOK_OK;
 }
@@ -164,7 +159,6 @@ static GGZHookReturn net_hook_failure(unsigned int id, const void *event_data,
 
 	error = *(GGZErrorEventData*)event_data;
 	printf("TelGGZ: Error %i: %s\n", id, error.message);
-	fflush(NULL);
 	exit(-1);
 
 	return GGZ_HOOK_OK;
@@ -195,7 +189,6 @@ static GGZHookReturn net_hook_roomlist(unsigned int id,
 				       const void *user_data)
 {
 	printf("TelGGZ: Logged in.\n");
-	fflush(NULL);
 
 	ggzcore_server_join_room(server, 0);
 
@@ -215,7 +208,6 @@ static GGZHookReturn net_hook_roomenter(unsigned int id,
 	{
 		flush_buffer();
 		printf("TelGGZ: %s has joined the room.\n", player);
-		fflush(NULL);
 	}
 	else
 	{
@@ -244,7 +236,6 @@ static GGZHookReturn net_hook_roomleave(unsigned int id,
 	{
 		flush_buffer();
 		printf("TelGGZ: %s has left the room.\n", player);
-		fflush(NULL);
 	}
 	else
 	{
@@ -274,7 +265,6 @@ static GGZHookReturn net_hook_chat(unsigned int id, const void *event_data,
 	{
 		flush_buffer();
 		printf("[%s]: %s\n", chat->sender, chat->message);
-		fflush(NULL);
 	}
 	else
 	{
@@ -305,7 +295,6 @@ static GGZHookReturn net_hook_prvmsg(unsigned int id, const void *event_data,
 	{
 		flush_buffer();
 		printf("[[%s]]: %s\n", player, message);
-		fflush(NULL);
 	}
 	else
 	{
@@ -329,9 +318,8 @@ static GGZHookReturn net_hook_enter(unsigned int id, const void *event_data,
 	room = ggzcore_server_get_cur_room(server);
 
 	printf("TelGGZ: Joined room %s.\n", ggzcore_room_get_name(room));
-	fflush(NULL);
 
-	if(!room)
+	if(room)
 	{
 		ggzcore_room_add_event_hook(room, GGZ_ROOM_ENTER, net_hook_roomenter);
 		ggzcore_room_add_event_hook(room, GGZ_ROOM_LEAVE, net_hook_roomleave);
